#ifndef ZAD2_QUAD_H
#define ZAD2_QUAD_H
#include <string>
#include <utility>
#include <cassert>

#include "../Info.h"
#include "Variable.h"
#include "Util.h"

using namespace std;
class Compiler;

static const std::string opNames[] = { "_U", "!", "", "-", "_UE",
                                       "_B", "+", "-", "*", "/", "%", "&", "|", "++", "_BE",
                                       "_C", "<", ">=", "<=", ">", "_CR", "_CQ", "==", "!=", "_CE"};

class Quadruple : public enable_shared_from_this<Quadruple> {
public:
	class Op {
	public:
		enum OpType {
			UNARY, // 0
			Not, Copy, Neg, UNARY_END, // 4
			BINARY, // 5
			Plus, Minus, Mul, Div, Mod, And, Or, Concat, BINARY_END, // 14
			CMP, // 15
			LT, GE, LE, GT, CMP_ROT_END, CMP_EQ, EQ, NEQ, CMP_END
		};
		OpType op;

		Op(OpType op = UNARY) : op(op) {};
		operator OpType() const { return op; };

		Op neg() const {
			if (op <= Op::Copy)
				return OpType(UNARY + 1 + ((int)(op - UNARY - 1)^1));
			if (op > Op::CMP)
				return OpType(CMP + 1 + ((int)(op - CMP - 1)^1));
			throw runtime_error("Invalid argument to Op::neg()");
		};

		Op rot() const {
			if (kind() != CMP) throw runtime_error("Invalid argument to Op::rot()");
			if (op < CMP_ROT_END) {
				// LT => GT, GE => LE, LE => GE, GT => LT
				return OpType((int)op ^ 3);
			}
			return op;
		}

		bool boolean() const {
			return kind() == CMP || op == Op::Not;
		}

		static OpType kind(OpType op) {
			if (op <= UNARY_END) return UNARY;
			if (op <= BINARY_END) return BINARY;
			return CMP;
		};

		OpType kind() const {
			return Op::kind(*this);
		};

		static const std::string& name(OpType op) {
			return opNames[op];
		}
	};

	int lineno;
	int index;
	BasicBlockPtr block;
	set<VariablePtr> aliveAfter;

	virtual ~Quadruple() = default;
	virtual std::string toString() const {  return to_string(lineno) + "\t"; };
	virtual bool isFinal() const { return false; }

	virtual void useVariables() {
		auto self = shared_from_this();
		for (const auto& v : vars()) {
			if (v && !v->constExpr) v->uses.emplace_back(self);
		}
		for (const auto& v : definitions()) {
			if (v) v->writes.emplace_back(self);
		}
	};
	virtual vector<VariablePtr> vars() const { return {}; };
	virtual vector<VariablePtr> definitions() const { return {}; };
	virtual void generateAsm(Compiler &c) { };
};

class QWriteVar : public Quadruple {
public:
	VariablePtr loc;
	explicit QWriteVar(VariablePtr loc) : Quadruple(), loc(std::move(loc)) {};

	std::string toString() const override { return Quadruple::toString() + "\t" + (loc ? (loc->name + " := ") : ""); }
	vector<VariablePtr> definitions() const override {
		auto ret = Quadruple::definitions();
		if (loc) ret.emplace_back(loc);
		return ret;
	};
};

class QLabel : public Quadruple {
public:
	string label;
	std::string toString() const override { return Quadruple::toString() + quote(label) + ":"; }
	explicit QLabel(std::string comment) : label(std::move(comment)) {}
	void generateAsm(Compiler &c) override;
};

class QAssign : public QWriteVar {
public:
	Op op;
	vector<VariablePtr> args;

	QAssign(VariablePtr loc, Op op, const vector<VariablePtr>& list) : QWriteVar(std::move(loc)), op(op) {
		for (const auto& i : list)
			if (i)
				args.emplace_back(i);
	};
	QAssign(VariablePtr loc, VariablePtr l, Op op, VariablePtr r) : QAssign(std::move(loc), op, {std::move(l), std::move(r)}) { };
	QAssign(VariablePtr loc, Op op, VariablePtr x) : QWriteVar(std::move(loc)), op(op), args({std::move(x)}) {};

	std::string toString() const override {
		if (args.size() == 1)
			return QWriteVar::toString() + Op::name(op) + " " + args[0]->name;
		else
			return QWriteVar::toString() + args[0]->name + " " + Op::name(op) + " " + args[1]->name;
	}

	vector<VariablePtr> vars() const override {
		return args;
	}
	void generateAsm(Compiler &c) override;
};

class QJump : public Quadruple {
public:
	shared_ptr<QLabel> target;
	explicit QJump(shared_ptr<QLabel> target) : target(std::move(target)) {};

	std::string toString() const override {
		return Quadruple::toString() + "jump " + quote(target->label);
	}

	bool isFinal() const override { return true; }
	void generateAsm(Compiler &c) override;
};

class QJumpCond : public QJump {
public:
	VariablePtr left;
	Op op;
	VariablePtr right;
	QJumpCond(shared_ptr<QLabel> to, VariablePtr left, Op op, VariablePtr right)
			: QJump(std::move(to)), left(std::move(left)), op(op), right(std::move(right)) {};

	QJumpCond(shared_ptr<QLabel> to, Op op, VariablePtr right)
			: QJumpCond(std::move(to), nullptr, op, std::move(right)) {};

	std::string toString() const override {
		if (!left)
			return QJump::toString() + " if " + Op::name(op) + " " + right->name;
		else
			return QJump::toString() + " if " + left->name + " " + Op::name(op) + " " + right->name;
	}

	bool isFinal() const override { return true; }

	vector<VariablePtr> vars() const override {
		auto ret = QJump::vars();
		if (left) ret.emplace_back(left);
		if (right) ret.emplace_back(right);
		return ret;
	}
	void generateAsm(Compiler &c) override;
};
class QCall;

class QParam : public Quadruple {
public:
	VariablePtr param;
	int num;

	shared_ptr<QCall> call;

	QParam(VariablePtr param, int num, shared_ptr<QCall> call = nullptr) : param(std::move(param)), num(num), call(std::move(call)) {};

	std::string toString() const override {
		if (num < 0) {
			return Quadruple::toString() + "\t" + param->name + " := param #" + to_string(-num);
		} else if (num == 0) {
			return Quadruple::toString() + "\tparam self " + param->name;
		} else {
			return Quadruple::toString() + "\tparam #" + to_string(num) + " <- " + param->name;
		}
	}

	vector<VariablePtr> vars() const override {
		auto ret = Quadruple::vars();
		if (param && call) ret.emplace_back(param);
		return ret;
	}

	vector<VariablePtr> definitions() const override {
		auto ret = Quadruple::definitions();
		if (param && !call) ret.emplace_back(param);
		return ret;
	};
	void generateAsm(Compiler &c) override;
};

class QCall : public QWriteVar {
public:
	FunctionInfoPtr fn;
	int params;
	VariablePtr self;
	QCall(VariablePtr loc, FunctionInfoPtr fn, int params, VariablePtr self)
		: QWriteVar(std::move(loc)), fn(std::move(fn)), params(params), self(std::move(self)) {};

	std::string toString() const override {
		if (self) {
			return QWriteVar::toString() + "call " + fn->name + " of " + self->name + " with " + to_string(params) + " params";
		} else {
			return QWriteVar::toString() + "call " + fn->name + " with " + to_string(params) + " params";
		}
	}

	vector<VariablePtr> vars() const override {
		auto ret = QWriteVar::vars();
		if (self) ret.emplace_back(self);
		return ret;
	}

	void generateAsm(Compiler &c) override;
};

class QReturn : public Quadruple {
public:
	VariablePtr val;
	explicit QReturn(VariablePtr ret) : val(std::move(ret)) {};

	std::string toString() const override {
		if (val) {
			return Quadruple::toString() + "\treturn " + val->name;
		} else {
			return Quadruple::toString() + "\treturn";
		}
	}

	bool isFinal() const override { return true; }

	vector<VariablePtr> vars() const override {
		auto ret = Quadruple::vars();
		if (val) ret.emplace_back(val);
		return ret;
	}
	void generateAsm(Compiler &c) override;
};

class QAccess : public QWriteVar {
public:
	QAccess(VariablePtr loc, VariablePtr base, VariablePtr index, int multiplier, int offset)
		: QWriteVar(std::move(loc)), access(std::move(base), std::move(index), multiplier, offset)
	{};
	// leal    8(,%eax,4), %eax         # Arithmetic: multiply eax by 4 and add 8
	// leal    (%edx,%eax,2), %eax      # Arithmetic: multiply eax by 2 and add edx
	VariableLocation access;

	std::string toString() const override {
		auto ret = QWriteVar::toString();
		return ret + access.toString();
	}

	vector<VariablePtr> vars() const override {
		auto ret = QWriteVar::vars();
		if (access.base) ret.emplace_back(access.base);
		if (access.index) ret.emplace_back(access.index);
		return ret;
	}
	void generateAsm(Compiler &c) override;
};

class QWrite : public Quadruple {
public:
	QWrite(VariableLocation loc, VariablePtr val)
		: Quadruple(), loc(std::move(loc)), val(std::move(val))
	{};
	VariableLocation loc;
	VariablePtr val;

	std::string toString() const override { return Quadruple::toString() + "\t" + loc.toString() + " := " + val->name; }

	vector<VariablePtr> vars() const override {
		auto ret = Quadruple::vars();
		if (loc.base) ret.emplace_back(loc.base);
		if (loc.index) ret.emplace_back(loc.index);
		if (val) ret.emplace_back(val);
		return ret;
	}
	void generateAsm(Compiler &c) override;
};

class QAlloc : public QWriteVar {
public:
	size_t size;
	VariablePtr count;
	std::string virtSymbol;
	QAlloc(VariablePtr loc, size_t size, std::string virtSymbol, VariablePtr count = nullptr)
		: QWriteVar(std::move(loc)), size(size), count(std::move(count)), virtSymbol(std::move(virtSymbol)) {};

	std::string toString() const override {
		std::string ret = QWriteVar::toString() + "alloc (" + to_string(size) + ")";
		if (count) { ret += " [" + count->name + "]"; }
		if (!virtSymbol.empty()) { ret += " " + virtSymbol; }
		return ret;
	}

	vector<VariablePtr> vars() const override {
		auto ret = QWriteVar::vars();
		if (count) ret.emplace_back(count);
		return ret;
	}
	void generateAsm(Compiler &c) override;
};

class QPhi : public Quadruple {
public:
	vector<map<VariablePtr, VariablePtr>> phi;

	vector<VariablePtr> vars() const override {
		auto ret = Quadruple::vars();
		for (const auto& b : phi) {
			for (const auto& v : b) {
				ret.emplace_back(v.second);
			}
		}
		return ret;
	}

	void useVariables() override {
		set<VariablePtr> written;
		auto self = shared_from_this();
		for (const auto& b : phi) {
			for (const auto &v : b) {
				if (!written.count(v.first)) {
					v.first->writes.emplace_back(self);
					written.emplace(v.first);
				}

				// QPhi does not place uses here, it is handled in RegisterAllocator
			}
		}
	}

	vector<VariablePtr> definitions() const override {
		auto ret = Quadruple::definitions();
		set<VariablePtr> written;
		for (const auto& b : phi) {
			for (const auto &v : b) {
				written.emplace(v.first);
			}
		}
		std::copy(written.begin(), written.end(), std::back_inserter(ret));
		return ret;
	};

	std::string toString() const override {
		if (phi.size() < 2) return "";
		std::string ret = Quadruple::toString() + "phi ";
		for (const auto& left : *phi.begin()) {
			auto &tg = left.first;
			ret += tg->name + " = [";
			bool f = true;
			for (auto b : phi) {
				if (!b.count(tg)) continue;
				if (f) f=false;
				else ret += ", ";
				ret += b[tg]->name;
			}
			ret += "] ";
		}
		return ret;
	}
	void generateAsm(Compiler &c) override;
};

class QBlockEnd : public Quadruple {
public:
	QBlockEnd(BasicBlockPtr b) : next(std::move(b)) {};
	QBlockEnd() : next(nullptr) {};
	BasicBlockPtr next;

	std::string toString() const override {
		return Quadruple::toString() + "phi_xchg";
	}
	void generateAsm(Compiler &c) override;
};


#endif //ZAD2_QUAD_H
