#include <cassert>
#include <iostream>
#include <filesystem>
#include "Quadruple.h"
#include "../Compiler.h"
#include "QuadrupleGenerator.h"

QuadrupleGenerator::Result QuadrupleGenerator::compileFunction(FunctionInfoPtr f) {
	if (!f->block) return QuadrupleGenerator::Result(strings);

	scope->currentFunction = f;
	newBlock();
	{
		int param = 1;
		if (VarInfoPtr self = f->self){
			auto var = alloc(self);
			addQuad(make_shared<QParam>(var, -param));
			param++;
		}
		for (const auto& arg : f->arguments) {
			auto var = alloc(arg);
			addQuad(make_shared<QParam>(var, -param));
			param++;
		}
	}
	f->block->accept(this);

	int id = 1;
	for (const auto& v : vars) {
		if (v->constExpr && !v->info) {
			v->name = to_string(v->val);
			continue;
		}
		v->name = "@" + to_string(id);
		if (v->info) {
			v->name += "." + v->info->name + ":" + to_string(v->info->lineLocation);
		}
		id++;
	}

	flushBasicBlock();
	block = nullptr;

	for (const auto& b : blocks) {
		b->finishQuads();
	}

	int label = 1;
	int index = 1;
	for (const auto& b : blocks) {
		for (const auto& q : b->quads) {
			q->index = index++;
			if (auto l = dynamic_pointer_cast<QLabel>(q)) {
				auto old = l->label;
				l->label = "." + to_string(label++) + "_" + Compiler::mangleFunctionName(f);
				if (!old.empty())
					l->label += "_" + old;
			}
		}
	}

	for (const auto& v : vars) {
		if (v->info)
			v->info->loc = nullptr;
	}
	QuadrupleGenerator::Result ret(strings);
	vars.swap(ret.vars);
	blocks.swap(ret.blocks);

	scope->currentFunction = nullptr;
	return ret;
}

BasicBlockPtr QuadrupleGenerator::flushBasicBlock(bool append) {
	auto ret = block;
	if (block) {
		blocks.emplace_back(block);
		flushLastQuad();
	}

	newBlock();

	if (append) {
		auto be = make_shared<QBlockEnd>(block);
		be->lineno = lineno;
		be->block = ret;
		ret->quads.emplace_back(be);
		ret->append(block);
	}

	return ret;
}

VariableLayout QuadrupleGenerator::captureEnv() {
	VariableLayout ret;
	BindingPtr b = scope->currentBinding;
	while (b) {
		for (const auto& info : b->variables) {
			if (info->isInstanceVariable()) continue;
			ret.add(info);
		}
		b = b->getParent();
	}
	return ret;
}

/// expressions

void QuadrupleGenerator::visitEVar(EVar *p) {
	auto var = p->pident_->getVar();
	lastVar = alloc(var);
	if (var->isInstanceVariable()) {
		auto fn = scope->currentFunction;
		auto self = fn->self;
		addQuad<QAccess>(lastVar, alloc(self), nullptr, 0, (int)var->offset);
	}
	Bool b;
	if (var->type->isEqual(&b)) {
		addLastVarCondJump(nullptr, Op::Copy, lastVar);
	}
}

void QuadrupleGenerator::visitELitInt(ELitInt *p) {
	lastVar = alloc(p->integer_);
}

void QuadrupleGenerator::visitELitTrue(ELitTrue *p) {
	lastVar = alloc(1);
}

void QuadrupleGenerator::visitELitFalse(ELitFalse *p) {
	lastVar = alloc(0);
}

void QuadrupleGenerator::visitEString(EString *p) {
	lastVar = alloc(getStringId(p->string_));
	lastVar->type = p->type;
}

void QuadrupleGenerator::visitNullCast(NullCast *p) {
	lastVar = alloc(0);
	lastVar->type = p->type;
}

void QuadrupleGenerator::visitNeg(Neg *p) {
	auto var = evalExpr(p->expr_);
	lastVar = alloc(var);
	addQuad<QAssign>(lastVar, Op::Neg, var);
	lastVar->val = -var->val;
}

void QuadrupleGenerator::visitNot(Not *p) {
	auto var = evalJump(p->expr_, labelFalse, labelTrue);
	lastVar = alloc(var);
	addQuad<QAssign>(lastVar, Op::Not, var);
	lastVar->val = !var->val;
}

void QuadrupleGenerator::visitEMul(EMul *p) {
	auto l = evalExpr(p->expr_1);
	auto r = evalExpr(p->expr_2);
	lastVar = alloc(l);
	p->mulop_->accept(this);
	addQuad<QAssign>(lastVar, l, op, r);
	if ((lastVar->constExpr = l->constExpr && r->constExpr)) {
		if (op == Op::Mul)
			lastVar->val = l->val * r->val;
		else if (r->val == 0)
			throw ParseError("Division by 0", p);
		else if (op == Op::Div)
			lastVar->val = l->val / r->val;
		else if (op == Op::Mod)
			lastVar->val = l->val % r->val;
	}
}

void QuadrupleGenerator::visitEAdd(EAdd *p) {
	auto l = evalExpr(p->expr_1);
	auto r = evalExpr(p->expr_2);
	lastVar = alloc(l);
	p->addop_->accept(this);
	Str s;
	if (p->type->isEqual(&s)) {
		auto call = make_shared<QCall>(lastVar, scope->functions.local("__latte_concat"), 2, nullptr);
		addQuad<QParam>(r, 1, call);
		addQuad<QParam>(l, 2, call);
		addQuad(call);
		lastVar->constExpr = false;
		return;
	}

	addQuad<QAssign>(lastVar, l, op, r);
	if ((lastVar->constExpr = l->constExpr && r->constExpr)) {
		if (op == Op::Minus)
			lastVar->val = l->val - r->val;
		else if (op == Op::Plus)
			lastVar->val = l->val + r->val;
	}
}

void QuadrupleGenerator::visitERel(ERel *p) {
	auto l = evalExpr(p->expr_1);
	auto r = evalExpr(p->expr_2);
	lastVar = alloc(l);
	p->relop_->accept(this);
	if ((lastVar->constExpr = l->constExpr && r->constExpr)) {
		if (op == Op::LT)
			lastVar->val = l->val < r->val;
		else if (op == Op::LE)
			lastVar->val = l->val <= r->val;
		else if (op == Op::EQ)
			lastVar->val = l->val == r->val;
		else if (op == Op::NEQ)
			lastVar->val = l->val != r->val;
		else if (op == Op::GE)
			lastVar->val = l->val >= r->val;
		else if (op == Op::GT)
			lastVar->val = l->val > r->val;
		addQuad<QAssign>(lastVar, l, op, r);
	} else {
		addLastVarCondJump(l, op, r);
	}
}

void QuadrupleGenerator::visitEAnd(EAnd *p) {
	auto labelAfter = make_shared<QLabel>("and_else");
	auto l = evalJump(p->expr_1, nullptr, labelFalse);
	if (!l) return;
	if (l->constExpr) {
		lastVar = l->val ? evalJump(p->expr_2, labelTrue, labelFalse) : l;
		// jest git skoczymy pozniej bo to constexpr
		return;
	} else {
		auto leftValBlock = block;
		auto llq = make_shared<QJumpCond>(labelFalse ? labelFalse : labelAfter, Op::Not, l);
		addQuad(llq);
		flushBasicBlock();
		leftValBlock->append(block);

		auto r = evalJump(p->expr_2, labelTrue, labelFalse);

		if (!r || r->constExpr) {
			assert(leftValBlock->quads.back() == llq);
			leftValBlock->quads.pop_back();
			// could merge basic blocks now, but why bother
			if (r) {
				lastVar = r->val ? l : r;
			}
		} else {
			auto rightValBlock = block;
			flushBasicBlock();
			addQuad(labelAfter);

			rightValBlock->append(block);
			leftValBlock->append(block);

			if (!labelTrue || !labelFalse) {
				lastVar = alloc();
				block->addPhi(rightValBlock, lastVar, r);
				block->addPhi(leftValBlock, lastVar, l);
			} else {
				lastVar = nullptr;
			}
		}
	}
}

void QuadrupleGenerator::visitEOr(EOr *p) {
	auto labelAfter = make_shared<QLabel>("or_else");
	auto l = evalJump(p->expr_1, labelTrue, nullptr);
	if (!l) return;
	if (l->constExpr) {
		lastVar = l->val ? l : evalJump(p->expr_2, labelTrue, labelFalse);
		// jest git skoczymy pozniej bo to constexpr
		return;
	} else {
		auto leftValBlock = block;
		auto llq = make_shared<QJumpCond>(labelTrue ? labelTrue : labelAfter, Op::Copy, l);
		addQuad(llq);
		flushBasicBlock();
		leftValBlock->append(block);

		auto r = evalJump(p->expr_2, labelTrue, labelFalse);

		if (!r || r->constExpr) {
			assert(leftValBlock->quads.back() == llq);
			leftValBlock->quads.pop_back();
			// could merge basic blocks now, but why bother
			if (r) {
				lastVar = r->val ? r : l;
			}
		} else {
			auto rightValBlock = block;
			flushBasicBlock();
			addQuad(labelAfter);

			rightValBlock->append(block);
			leftValBlock->append(block);

			if (!labelTrue || !labelFalse) {
				lastVar = alloc();
				block->addPhi(rightValBlock, lastVar, r);
				block->addPhi(leftValBlock, lastVar, l);
			} else {
				lastVar = nullptr;
			}
		}
	}
}

/// complex extensions

void QuadrupleGenerator::visitEIndexAcc(EIndexAcc *p) {
	auto lhs = evalExpr(p->expr_1);
	auto index = evalExpr(p->expr_2);
	auto type = dynamic_pointer_cast<Array>(lhs->type);
	if (!type) {
		throw ParseError("compiler evalExpr Variable has no Array type", p->expr_1);
	}
	if (!type->type_) {
		throw ParseError("compiler evalExpr Variable is Array of no type", p->expr_1);
	}
	lastVar = alloc(type->type_->clone());
	lastVar->memory = true;
	assert(type->type_->size() <= 4);
	addQuad<QAccess>(lastVar, lhs, index, type->type_->size(), 4); // 4 accounts for table length entry
	Bool b;
	if (type->type_->isEqual(&b)) {
		addLastVarCondJump(nullptr, Op::Copy, lastVar);
	}
}

void QuadrupleGenerator::visitEClsMmbr(EClsMmbr *p) {
	auto l = evalJump(p->expr_, labelTrue, labelFalse);
	auto var = p->pident_->getVar();
	assert(var);
	// it cannot be function, as function might only be used in EApp and are handled directly there
	assert(!var->toFunction());
	// Array length calculation works because of arrayLengthVar that has offset 0
	lastVar = alloc(var);
	addQuad<QAccess>(lastVar, l, nullptr, 0, var->offset);
	Bool b;
	if (var->type->isEqual(&b)) {
		addLastVarCondJump(nullptr, Op::Copy, lastVar);
	}
}

void QuadrupleGenerator::visitEApp(EApp *p) {
	auto funType = static_pointer_cast<Fun>(p->expr_->type);
	FunctionInfoPtr info;
	VariablePtr self;
	if (auto mem = dynamic_cast<EClsMmbr*>(p->expr_)) {
		// call tab[i].fun()
		// call obj.fun()
		self = evalExpr(mem->expr_);
		info = static_pointer_cast<FunctionInfo>(mem->pident_->getVar());
	} else if (auto ident = dynamic_cast<EVar*>(p->expr_)) {
		auto klass = scope->currentFunction->klass.lock();
		info = static_pointer_cast<FunctionInfo>(ident->pident_->getVar());
		auto otherClass = info->klass.lock();
		if (otherClass) {
			// call fun() = self.fun()
			assert(otherClass == klass);
			auto binding = scope->currentFunction->binding.lock();
			assert(binding);
			auto selfInfo = binding->variables.local("self");
			assert(selfInfo);
			self = alloc(selfInfo);
		} else {
			// call main()
		}
	} else {
		throw ParseError("Unimplemented EApp instantiation (neither EClsMmbr nor EVar", p);
	}

	auto ret = alloc(funType->type_->clone());
	auto call = make_shared<QCall>(ret, info, p->listexpr_->size(), self);
	int i = 0;
	for (auto param = p->listexpr_->rbegin(); param != p->listexpr_->rend(); param++) {
		auto var = evalExpr(*param);
		addQuad<QParam>(var, ++i, call);
	}

	if (self) {
		addQuad<QParam>(self, 0, call);
	}

	lastVar = ret;
	addQuad(call);

	Bool b;
	if (ret->type->isEqual(&b)) {
		addLastVarCondJump(nullptr, Op::Copy, lastVar);
	}
}

void QuadrupleGenerator::visitENewArray(ENewArray *p) {
	// allocate enough memory with calloc, if it is class then init each entry...
	auto type = p->type_;
	size_t size = type->size();
	auto count = evalExpr(p->expr_);
	lastVar = alloc(new Array(type->clone()));
	addQuad<QAlloc>(lastVar, size, "", count);
}

void QuadrupleGenerator::visitENewClass(ENewClass *p) {
	// allocate enough memory with calloc, setup virt table pointer.
	auto info = p->pident_->getVar()->toClass();
	size_t size = info->size;
	string virtSymbol = Compiler::getVirtName(info);
	lastVar = alloc();
	addQuad<QAlloc>(lastVar, size, virtSymbol);
}

/// statements

void QuadrupleGenerator::visitInit(Init *p) {
	auto info = p->pident_->getVar();
	assert(info);
	auto var = evalExpr(p->expr_);
	auto loc = alloc(info);
	addQuad<QAssign>(loc, Op::Copy, var);
	if ((loc->constExpr = var->constExpr))
		loc->val = var->val;
}

void QuadrupleGenerator::visitNoInit(Init *p) {
	auto info = p->pident_->getVar();
	assert(info);
	auto loc = alloc(info);
	addQuad<QAssign>(loc, Op::Copy, alloc(0));
	loc->constExpr = true;
	loc->val = 0;
}

void QuadrupleGenerator::assign(Expr* lval, VariablePtr val) {
	auto dest = evalLVal(lval);
	if (dest->isInstanceVariable()) {
		// instance variable, need to write it to memory
		auto quad = dynamic_pointer_cast<QAccess>(lastQuad);
        assert(quad);
		auto loc = quad->access;
		lastQuad = nullptr;
		addQuad<QWrite>(loc, val);
	} else {
		// local variable - only assign
		addQuad<QAssign>(dest, Op::Copy, val);
		if ((dest->constExpr = val->constExpr))
			dest->val = val->val;
	}
}

void QuadrupleGenerator::visitAss(Ass *p) {
	auto val = evalExpr(p->expr_2);
	assign(p->expr_1, val);
}

void QuadrupleGenerator::visitIncr(Incr *p) {
	auto lhs = evalExpr(p->expr_);
	auto tmp = alloc();
	addQuad<QAssign>(tmp, lhs, Op::Plus, alloc(1));
	assign(p->expr_, tmp);
}

void QuadrupleGenerator::visitDecr(Decr *p) {
	auto lhs = evalExpr(p->expr_);
	auto tmp = alloc();
	addQuad<QAssign>(tmp, lhs, Op::Minus, alloc(1));
	assign(p->expr_, tmp);
}

void QuadrupleGenerator::visitBlk(Blk *blk) {
	scope->currentBinding = blk->getBinding();

	Skeleton::visitBlk(blk);

	scope->currentBinding = scope->currentBinding->getParent();
}

/// flow control

void QuadrupleGenerator::visitRet(Ret *p) {
	auto var = evalExpr(p->expr_);
	addQuad<QReturn>(var);
	flushBasicBlock();
}

void QuadrupleGenerator::visitVRet(VRet *p) {
	addQuad<QReturn>(nullptr);
	flushBasicBlock();
}

void QuadrupleGenerator::visitCond(Cond *p) {
	compileCond(p->expr_, p->stmt_, nullptr);
}

void QuadrupleGenerator::visitCondElse(CondElse *p) {
	compileCond(p->expr_, p->stmt_1, p->stmt_2);
}

void QuadrupleGenerator::compileCond(Expr *expr_, Stmt *stmt_1, Stmt *stmt_2) {
	auto elseBranch = make_shared<QLabel>("if_else");
	auto after = make_shared<QLabel>("end_else");
	auto ifBranch = make_shared<QLabel>("if_true");

	auto condVar = evalJump(expr_, ifBranch, elseBranch);
	if (condVar->constExpr) {
		assert(!dynamic_pointer_cast<QJump>(lastQuad));
		if (condVar->val) {
			addQuad(ifBranch);
			stmt_1->accept(this);
		} else if (stmt_2) {
			addQuad(elseBranch);
			stmt_2->accept(this);
		}
		return;
	}

	auto beforeBlock = flushBasicBlock(); // possible jump -> else
	auto envIf = captureEnv(), envElse = envIf;
	auto stmtIfFirstBlock = block;
	addQuad(ifBranch);
	stmt_1->accept(this);
	addQuad<QJump>(after);
	auto stmtIfLastBlock = flushBasicBlock(); // jump -> after
	envIf.capture();
	envIf.revert();

	addQuad(elseBranch);
	auto stmtElseFirstBlock = block;
	if (stmt_2) {
		stmt_2->accept(this);
	}

	auto stmtElseLastBlock = flushBasicBlock(true); // jump <- cond
	envElse.capture();
	addQuad(after);

	beforeBlock->append(stmtIfFirstBlock);
	beforeBlock->append(stmtElseFirstBlock);
	stmtIfLastBlock->append(block);

	merge2Envs(&envIf, stmtIfLastBlock, &envElse, stmtElseLastBlock);
}

void QuadrupleGenerator::merge2Envs(VariableLayout *env1, const BasicBlockPtr& b1, VariableLayout *env2, const BasicBlockPtr& b2) {
	assert(env1->changes.size() == env2->changes.size());
	for (auto it = env1->changes.begin(), itt = env2->changes.begin(); it != env1->changes.end(); it++, itt++) {
		assert(it->first == itt->first);
		assert(it->second.first == itt->second.first);
		if (it->second.second == itt->second.second) continue;
		auto info = it->first;
		info->loc = nullptr;
		auto merged = alloc(info);
		block->addPhi(b1, merged, it->second.second);
		block->addPhi(b2, merged, itt->second.second);
	}
}

void QuadrupleGenerator::visitWhile(While *expr) {
	auto cond = make_shared<QLabel>("cond");

	addQuad<QJump>(cond);
	auto beforeBlock = flushBasicBlock(); // jump <- loop -> cond
	auto loop = make_shared<QLabel>("loop");
	addQuad(loop); // jump <- loop

	// hook env
	auto env1 = captureEnv();

	// produce all variables hooks
	for (const auto& change : env1.changes) {
		if (!change.second.first) continue;
		auto info = change.first;
		info->loc = nullptr;
		alloc(info);
	}

	// save hooks for later
	env1.capture();

	// env2 starts with hooked variables
	auto env2 = captureEnv();
	auto loopFirstBlock = block;

	expr->stmt_->accept(this);

	auto loopLastBlock = flushBasicBlock(true); // jump <- cond
	auto condFirstBlock = block;
	addQuad(cond);

	// env2 contains changed hooks
	env2.capture();
	// restore env1 pre-hook variables
	env1.revert();

	beforeBlock->append(condFirstBlock);

	for (const auto& p : env2.changes) {
		auto info = p.first;
		auto hooked = p.second.first;
		auto looped = p.second.second;
		if (!hooked) continue;
		auto orig = env1.changes[info].first;
		// save hooks if used
		info->loc = hooked;
		// transition from pre-hook to hooked var [before -> cond]
		condFirstBlock->addPhi(beforeBlock, hooked, orig);
		// transition from hooked var
		condFirstBlock->addPhi(loopLastBlock, hooked, looped);

		// loopLastBlock doesn't need phi, it has only one incoming block
	}

	// expr uses pre-hook variables iff unused in loop
	auto var = evalJump(expr->expr_, loop, nullptr);
	auto condLastBlock = flushBasicBlock(true); // jump -> loop

	condLastBlock->append(loopFirstBlock);

	// next block is ready to use updated variables
}

void QuadrupleGenerator::visitSExp(SExp *p) {
	evalExpr(p->expr_);
}

void QuadrupleGenerator::visitForEach(ForEach *expr) {
	auto cond = make_shared<QLabel>("for_cond");

	auto arr = evalExpr(expr->expr_);
	addQuad<QJump>(cond);
	auto beforeBlock = flushBasicBlock(); // jump <- loop -> cond
	auto loop = make_shared<QLabel>("for_loop");
	addQuad(loop); // jump <- loop

	// hook env
	auto env1 = captureEnv();

	// produce all variables hooks
	for (const auto& change : env1.changes) {
		if (!change.second.first) continue;
		auto info = change.first;
		info->loc = nullptr;
		alloc(info);
	}

	VarInfoPtr iter = make_shared<VarInfo>("_it", make_shared<Int>());
	auto iter1 = alloc(iter);

	// save hooks for later
	env1.capture();

	// env2 starts with hooked variables
	auto env2 = captureEnv();
	auto loopFirstBlock = block;

	auto iterInfo = expr->pident_->getVar();
	assert(iterInfo);
	auto iterVar = alloc(iterInfo);
	addQuad<QAccess>(iterVar, arr, iter1, 4, 4);

	expr->stmt_->accept(this);

	iter->loc = nullptr;
	auto iter2 = alloc(iter);
	addQuad<QAssign>(iter2, iter1, Op::Plus, alloc(1));

	auto loopLastBlock = flushBasicBlock(true); // jump <- cond
	auto condFirstBlock = block;
	addQuad(cond);

	// env2 contains changed hooks
	env2.capture();
	// restore env1 pre-hook variables
	env1.revert();

	beforeBlock->append(condFirstBlock);

	for (const auto& p : env2.changes) {
		auto info = p.first;
		auto hooked = p.second.first;
		auto looped = p.second.second;
		if (!hooked) continue;
		auto orig = env1.changes[info].first;
		// save hooks if used
		info->loc = hooked;
		// transition from pre-hook to hooked var [before -> cond]
		condFirstBlock->addPhi(beforeBlock, hooked, orig);
		// transition from hooked var
		condFirstBlock->addPhi(loopLastBlock, hooked, looped);

		// loopLastBlock doesn't need phi, it has only one incoming block
	}

	condFirstBlock->addPhi(beforeBlock, iter1, alloc(0));
	condFirstBlock->addPhi(loopLastBlock, iter1, iter2);

	auto len = alloc();
	addQuad<QAccess>(len, arr, nullptr, 0, 0);
	addQuad<QJumpCond>(loop, iter1, Op::LT, len);
	auto condLastBlock = flushBasicBlock(true); // jump -> loop

	condLastBlock->append(loopFirstBlock);

	// next block is ready to use updated variables
}
